import logging
from enum import Enum

LOGGER = logging.getLogger(__name__)


class UserAttribute(Enum):

    UID = ('uid', 'username')
    GIVENNAME = ('givenName', 'givenname')
    SURNAME = ('sn', 'surname')
    EMAIL = ('mail', 'email')
    COUNTRYCODE = ('st', 'countrycode')
    COMPANY = ('MEPUserCompany', 'company')
    USER_USAGE = ('MEPUserUsageFields', 'user_usage')
    USER_CATEGORY = ('MEPUserCategory', 'user_category')
    USER_FUNCTION = ('MEPUserFunction', 'user_function')
    REGISTRATION_DATE = ('createTimestamp', 'registration_date')

    def __init__(self, ldap_attribute, db_column):
        self.ldap_attribute = ldap_attribute
        self.db_column = db_column


class User(object):

    def __init__(self, username, givenname, surname,
                 email, countrycode, company,
                 user_usage, user_category, user_function,
                 registration_date):

        self.username = username
        self.givenname = givenname
        self.surname = surname
        self.email = email
        self.countrycode = countrycode
        self.company = company
        self.user_usage = user_usage
        self.user_category = user_category
        self.user_function = user_function
        self.registration_date = registration_date

    def is_natural_person(self):

        return  self.company is not None \
                and self.user_usage is not None \
                and self.user_category is not None \
                and self.user_function is not None


class UserExporter(object):

    def __init__(self, ldap_url, ldap_username, ldap_password):
        """
        Checks if user is a natural person who registered using the web portal
        this filters out admin users and service users
        :param ldap_url: LDAP URL, e.g. ldap://ipa01.vgt.vito.be:389
        :param ldap_username: user to connect to LDAP service
        :param ldap_password: password for LDAP user
        """
        self.ldap_url = ldap_url
        self.ldap_username = ldap_username
        self.ldap_password = ldap_password

    @staticmethod
    def _get_user_attribute(user, attribute):
        return None if attribute not in user else user[attribute][0].decode('utf8')

    def _get_users_ldap(self):

        conn = None

        try:

            import ldap
            conn = ldap.initialize(self.ldap_url)
            conn.simple_bind_s(self.ldap_username, self.ldap_password)
            attributes = [UserAttribute.UID.ldap_attribute, UserAttribute.EMAIL.ldap_attribute, UserAttribute.GIVENNAME.ldap_attribute,
                          UserAttribute.SURNAME.ldap_attribute, UserAttribute.COUNTRYCODE.ldap_attribute, UserAttribute.COMPANY.ldap_attribute,
                          UserAttribute.USER_CATEGORY.ldap_attribute, UserAttribute.USER_FUNCTION.ldap_attribute, UserAttribute.USER_USAGE.ldap_attribute,
                          UserAttribute.REGISTRATION_DATE.ldap_attribute]
            results = conn.search_s('cn=users,cn=accounts,dc=vgt,dc=vito,dc=be',
                                 scope=ldap.SCOPE_ONELEVEL,
                                 filterstr='(objectClass=MEPUser)',
                                 attrlist=attributes)

            users = []
            for result in results:
                from datetime import datetime
                registration_date = datetime.strptime(
                    UserExporter._get_user_attribute(result[1], UserAttribute.REGISTRATION_DATE.ldap_attribute),
                    '%Y%m%d%H%M%SZ')
                user = User(UserExporter._get_user_attribute(result[1], UserAttribute.UID.ldap_attribute),
                            UserExporter._get_user_attribute(result[1], UserAttribute.GIVENNAME.ldap_attribute),
                            UserExporter._get_user_attribute(result[1], UserAttribute.SURNAME.ldap_attribute),
                            UserExporter._get_user_attribute(result[1], UserAttribute.EMAIL.ldap_attribute),
                            UserExporter._get_user_attribute(result[1], UserAttribute.COUNTRYCODE.ldap_attribute),
                            UserExporter._get_user_attribute(result[1], UserAttribute.COMPANY.ldap_attribute),
                            UserExporter._get_user_attribute(result[1], UserAttribute.USER_USAGE.ldap_attribute),
                            UserExporter._get_user_attribute(result[1], UserAttribute.USER_CATEGORY.ldap_attribute),
                            UserExporter._get_user_attribute(result[1], UserAttribute.USER_FUNCTION.ldap_attribute),
                            registration_date)
                if user.is_natural_person():
                    users.append(user)

            return users

        finally:
            conn.unbind()

    def export_csv(self, csv_file):
        """
        Export IPA users to CSV file.
        :param csv_file: CSV file containing exported users
        """

        users = self._get_users_ldap()

        with open(csv_file, 'w') as users_file:
            import csv
            users_writer = csv.writer(users_file, delimiter=',')
            for user in users:
                users_writer.writerow(
                    [user.username, user.givenname, user.surname, user.email, user.countrycode, user.company,
                     user.user_usage, user.user_category, user.user_function, user.registration_date])

        LOGGER.info(f'Exported {len(users)} users to file {users_file.name}')

    def export_db(self, db_host, db_name, db_username, db_password):
        """
        Export IPA users to a database
        :param db_host:
        :param db_name:
        :param db_username:
        :param db_password:
        """

        import psycopg2
        from psycopg2 import sql

        users = self._get_users_ldap()

        conn = None
        cursor = None
        try:
            conn = psycopg2.connect("host='{}' dbname='{}' user='{}' password='{}'"
                                    .format(db_host, db_name, db_username, db_password))
            cursor = conn.cursor()

            for user in users:
                cursor.execute(
                    sql.SQL(
                        """INSERT INTO oscars_users (username, email, givenname, surname, countrycode, company, 
                                                     user_usage, user_category, user_function, registration_date) 
                            VALUES (%s, %s, %s, %s, %s, %s, %s, %s, %s, %s)
                            ON CONFLICT ON CONSTRAINT oscars_users_username_key DO
                            UPDATE set email = EXCLUDED.email, 
                                       givenname = EXCLUDED.givenname,
                                       surname = EXCLUDED.surname,
                                       countrycode = EXCLUDED.countrycode,
                                       company = EXCLUDED.company,
                                       user_usage = EXCLUDED.user_usage,
                                       user_category = EXCLUDED.user_category,
                                       user_function = EXCLUDED.user_function"""),
                               [user.username, user.email, user.givenname, user.surname, user.countrycode,
                                user.company, user.user_usage, user.user_category, user.user_function, user.registration_date])
            conn.commit()
        except Exception as ex:
            LOGGER.error(f'Could not connect to database: {ex}')
            raise
        finally:
            if cursor is not None:
                cursor.close()
            if conn is not None:
                conn.close()
